MOVE_TECH_CHANCE_PERCENT_PER_FIELD = 30

--[[
	For every field, there is a given chance (MOVE_TECH_CHANCE_PERCENT_PER_FIELD) that one of the techs in 
	that field will be moved either one field up or one field down within the tech tree. 
	Note that when a tech is moved to a new field its RP cost is that of the new field, meaning 
	the tech becomes cheaper or more expensive to research. 
	
	The choice between moving up or down along the techtree is randomly determined (50:50) 
	separately for every tech that is being moved. If one of these two directions is not allowed, 
	the other direction is automatically used. If neither is allowed or the tech is the only tech 
	in its field, the tech is not moved. 
	
	A move-direction can be invalid for multiple reasons:
	- Fields which give the player all of the tech when researched (e.g. Colony Ship, Outpost Ship, Transport) 
	  are not affected meaning no tech is moved in or out of these fields. 
	- Hyper-advanced fields are not affected. 
	- If a field already has 4 techs in it, no other tech is moved into it. 
	- It is prevented to have two tech in the same field where one choice is clearly better, 
	  e.g. Fusion Drive and Ion Drive will never appear in the same field. This rule only applies 
	  if one tech is better in every aspect than the other, making it obsolete. A lower tier weapon 
	  or system might still be viable due to lower space or cost meaning the rule will not be applied. 
	  To see which tech are mutually exclusive, check the tech_groups variable below, all tech that are 
	  assigned the same number will not be paired up in a single research field. If the assigned number 
	  is 0 then they are excluded from this rule. 
	
	Note: If you plan to start a new game with this mod active, it is recommended to restart Orion 2 first. 
	The reason for this that any previously started or loaded game might have a different tech tree, which 
	is not reset by this mod. 
	
	For this mod, some code was reused from the RANDTECH mod of 1.50.18.
	
	Last update of this mod: 20th February 2021; Tested with version 1.50.13 improved
	
	Questions etc. to: romanhable@web.de (Epirasque)
]]

CATEGORIES_TO_EXCLUDE = {
	starting_tech, 
	xenon_technology,
	advanced_governments
}

tech_groups = {
	tachyon_communications = 1,
	sub_space_communications = 1,
	hyperspace_communications = 1,
	
	laser_rifle = 2,
	fusion_rifle = 2,
	phasor_rifle = 2,
	plasma_rifle = 2,
	pulse_rifle = 2,
	
	jump_gate = 3,
	star_gate = 3,
	
	nuclear_drive = 4,
	fusion_drive = 4,
	ion_drive = 4,
	anti_matter_drive = 4,
	hyper_drive = 4,
	interphased_drive = 4,
	
	space_scanner = 5,
	tachyon_scanner = 5,
	neutron_scanner = 5,
	sensors = 5,
	
	electronic_computer = 6,
	optronic_computer = 6,
	positronic_computer = 6,
	cybertronic_computer = 6,
	moleculartronic_computer = 6,
	
	titanium_armor = 7,
	tritanium_armor = 7,
	zortrium_armor = 7,
	neutronium_armor = 7,
	adamantium_armor = 7,
	xentronium_armor = 7,
	
	class_i_shield = 8,
	class_iii_shield = 8,
	class_v_shield = 8,
	class_vii_shield = 8,
	class_x_shield = 8,
	
	standard_fuel_cells = 9,
	deuterium_fuel_cells = 9,
	iridium_fuel_cells = 9,
	uridium_fuel_cells = 9,
	thorium_fuel_cells = 9,
	
	hydroponic_farm = 0,
	soil_enrichment = 0,
	terraforming = 0,
	subterranean_farms = 0,
	weather_control_system = 0,
	food_replicators = 0,
	android_farmers = 0,
	biomorphic_fungi = 0,
	automated_factories = 0,
	robo_miners = 0,
	planetary_gravity_generator = 0,
	recyclotron = 0,
	robotic_factory = 0,
	android_workers = 0,
	deep_core_mining = 0,
	research_laboratory = 0,
	planetary_supercomputer = 0,
	autolab = 0,
	heightened_intelligence = 0,
	android_scientists = 0,
	galactic_cybernet = 0,
	spaceport = 0,
	planetary_stock_exchange = 0,
	galactic_currency_exchange = 0,
	pollution_processor = 0,
	atmospheric_renewer = 0,
	nano_disassemblers = 0,
	core_waste_dumps = 0,
	space_academy = 0,
	alien_management_center = 0,
	astro_university = 0,
	warp_interdictor = 0,
	dimensional_portal = 0,
	capitol = 0,
	cloning_center = 0,
	microbiotics = 0,
	universal_antidote = 0,
	biospheres = 0,
	advanced_city_planning = 0,
	gaia_transformation = 0,
	marine_barracks = 0,
	armor_barracks = 0,
	colony_base = 0,
	star_base = 0,
	outpost_ship = 0,
	planetary_missile_base = 0,
	fighter_garrison = 0,
	battlestation = 0,
	ground_batteries = 0,
	star_fortress = 0,
	artemis_system_net = 0,
	stellar_converter = 0,
	planetary_radiation_shield = 0,
	planetary_flux_shield = 0,
	planetary_barrier_shield = 0,
	xeno_psychology = 0,
	planet_construction = 0,
	evolutionary_mutation = 0,
	neural_scanner = 0,
	telepathic_training = 0,
	cyber_security_link = 0,
	stealth_suit = 0,
	psionics = 0,
	spy_network = 0,
	confederation = 0,
	federation = 0,
	galactic_unification = 0,
	imperium = 0,
	anti_grav_harness = 0,
	powered_armor = 0,
	battleoids = 0,
	personal_shield = 0,
	battle_pods = 0,
	survival_pods = 0,
	advanced_damage_control = 0,
	microlite_construction = 0,
	megafluxers = 0,
	time_warp_facilitator = 0,
	quantum_detonator = 0,
	nuclear_bomb = 0,
	fusion_bomb = 0,
	anti_matter_bomb = 0,
	neutronium_bomb = 0,
	death_spores = 0,
	bio_terminator = 0,
	nuclear_missile = 0,
	merculite_missile = 0,
	pulson_missile = 0,
	anti_matter_torpedo = 0,
	proton_torpedo = 0,
	zeon_missile = 0,
	plasma_torpedo = 0,
	freighters = 0,
	colony_ship = 0,
	transport = 0,
	titan_construction = 0,
	doom_star_construction = 0,
	augmented_engines = 0,
	inertial_stabilizer = 0,
	inertial_nullifier = 0,
	laser_cannon = 0,
	fusion_beam = 0,
	mass_driver = 0,
	ion_pulse_cannon = 0,
	neutron_blaster = 0,
	graviton_beam = 0,
	phasor = 0,
	gauss_cannon = 0,
	plasma_cannon = 0,
	disrupter_cannon = 0,
	death_ray = 0,
	particle_beam = 0,
	battle_scanner = 0,
	rangemaster_unit = 0,
	structural_analyzer = 0,
	high_energy_focus = 0,
	hyper_x_capacitors = 0,
	achilles_targeting_unit = 0,
	anti_missile_rocket = 0,
	ecm_jammer = 0,
	lightning_field = 0,
	pulsar = 0,
	multi_wave_ecm_jammer = 0,
	wide_area_jammer = 0,
	spatial_compressor = 0,
	dauntless_guidance_system = 0,
	fast_missile_racks = 0,
	emissions_guidance_system = 0,
	fighter_bays = 0,
	bomber_bays = 0,
	heavy_fighter_bays = 0,
	reinforced_hull = 0,
	heavy_armor = 0,
	automated_repair_unit = 0,
	damper_field = 0,
	shield_capacitors = 0,
	multi_phased_shields = 0,
	hard_shields = 0,
	reflection_field = 0,
	troop_pods = 0,
	assault_shuttles = 0,
	tractor_beam = 0,
	transporters = 0,
	extended_fuel_tanks = 0,
	scout_lab = 0,
	security_stations = 0,
	warp_dissipator = 0,
	sub_space_teleporter = 0,
	stealth_field = 0,
	cloaking_device = 0,
	displacement_device = 0,
	phasing_cloak = 0,
	gyro_destabilizer = 0,
	stasis_field = 0,
	energy_absorber = 0,
	plasma_web = 0,
	mauler_device = 0,
	black_hole_generator = 0,
	phase_shifter = 0,
	holo_simulator = 0,
	virtual_reality_network = 0,
	pleasure_dome = 0
}

advf = {
    advf_biology = 1,
    advf_power = 1,
    advf_physics = 1,
    advf_construction = 1,
    advf_fields = 1,
    advf_chemistry = 1,
    advf_computers = 1,
    advf_sociology = 1
}

gray_box_field_names = {
	nuclear_fission = 1, 
	cold_fusion = 1,
	chemistry = 1,
	electronics = 1,
	physics = 1,
	engineering = 1
}

starting_fields = rget_starting_fields()
PREVIOUS = 1
NEXT = 2

moved_techs = {}

function set_id(t, tag)
    for i,v in pairs(t) do
        v[tag] = i
    end
end

function make_branches(ff, keep_gray_boxes)
    local branches = {}
    local f2br = {}

    local map = { engineering = "advanced_engineering" }
    if keep_gray_boxes then
        map.chemistry = "advanced_metallurgy"
        map.nuclear_fission = "advanced_fusion"
        map.physics = "fusion_physics"
        map.electronics = "optronics"
    end
    for i,s in pairs(starting_fields) do
        if map[s] ~= nil then
            starting_fields[i] = map[s]
        end
    end

    for i,f in pairs(starting_fields) do
        local n = rawlen(branches) + 1
        local b = {
            branch_last_field = "advf_" .. i,
            branchnum = n,
            field_id16 = f,
        }
        table.insert(branches, b)
        ff[f].is_branch_root = 1
    end

    for i,b in ipairs(branches) do
        local fname = b.field_id16
        local marks = {}
        local n = 1
        while advf[fname] == nil and ff[fname] ~= nil do
            local f = ff[fname]
            if marks[fname] ~= nil then
                error(tostring("detected tech loop ", marks))
            end
            marks[fname] = n
            n = n + 1
            f.branch = i
            fname = f.next_field
        end
    end

    return branches
end

function add_tech_names_to_fields(techs, fields)
	for tech_name, tech in pairs(techs) do
		local field = fields[tech.field_id16]
		if field ~= nil then
			if field.techs == nil then
				field.techs = {}
			end
			table.insert(field.techs, tech_name)
		end
	end
	return fields
end

function is_excluded_field(field_to_check, tech_to_move)
	local field_name = field_to_check.name
	local techs_to_check = field_to_check.techs
	local excluded = false
	for _, tech_name in pairs(techs_to_check) do
		if tech_to_move ~= nil and tech_groups[tech_to_move] ~= nil and tech_groups[tech_to_move] > 0 and tech_groups[tech_to_move] == tech_groups[tech_name] then
			excluded = true
		end
	end
	if field_name == nil or CATEGORIES_TO_EXCLUDE[field_name] ~= nil or advf[field_name] ~= nil or starting_fields[field_name] ~= nil or gray_box_field_names[field_name] ~= nil then
		excluded = true
	end
	return excluded
end

function move_techs(tt, fields)
	for field_name, field in pairs(fields) do
		if (not is_excluded_field(field, nil)) and random(100) <= MOVE_TECH_CHANCE_PERCENT_PER_FIELD then
			local unfiltered_tech_candidates = field.techs
			local tech_candidates = {}
			for candidate_id, candidate_name in pairs(unfiltered_tech_candidates) do
				if moved_techs[candidate_name] == nil then
					table.insert(tech_candidates, candidate_name)
				else
				end
			end
			
			local valid_to_move = true
			local tech_to_move_id = nil
			local tech_to_move = nil
			if #tech_candidates > 0 then
				tech_to_move_id = random(#tech_candidates)
				tech_to_move = tech_candidates[tech_to_move_id]
			else
				valid_to_move = false
			end
			
			if #field.techs == 1 then
				valid_to_move = false
			end
			move_direction = random(2)
			if is_excluded_field(fields[field.prev_field], tech_to_move) and is_excluded_field(fields[field.next_field], tech_to_move) then
				valid_to_move = false
			elseif is_excluded_field(fields[field.prev_field], tech_to_move) then
				move_direction = NEXT
			elseif is_excluded_field(fields[field.next_field], tech_to_move) then
				move_direction = PREVIOUS
			end
			field_to_move_to = nil
			if move_direction == NEXT then
				field_to_move_to = fields[field.next_field]
			elseif move_direction == PREVIOUS then
				field_to_move_to = fields[field.prev_field]
			end
			if field_to_move_to ~= nil and #field_to_move_to.techs == 4 then
				valid_to_move = false
			end
			if valid_to_move then
				tt[tech_to_move].field_id16 = field_to_move_to.name
				table.remove(field.techs, tech_to_move_id)
				table.insert(field_to_move_to.techs, tech_to_move)
				moved_techs[tech_to_move] = true
				nr_moved_techs = nr_moved_techs + 1
			end
		end
	end
	return tt
end

function go()
    set_seed()
	
    local ff = rget_fields() set_id(ff, "name")
    local tt = rget_techs()  set_id(tt, "name")
	local bb = make_branches(ff, 1)
	ff = add_tech_names_to_fields(tt, ff)

	tt = move_techs(tt, ff)
	
    rset_fields(ff)
    rset_techs(tt)

	msgbox(nr_moved_techs .. " tech were moved up or down one field in the tech tree (" .. MOVE_TECH_CHANCE_PERCENT_PER_FIELD .. "% chance for a field to move one tech)")

end

nr_moved_techs = 0

go()
